import os
import json
import glob
import pandas as pd
from scipy.io import loadmat    
import xml.etree.ElementTree as ET

from .config import phases_dict as default_phases_dict
from .config import phase_grouping as default_phase_grouping

__author__ = "Miguel Ponce de Leon"
__copyright__ = "Copyright 2020, PhysiCell ToolKit project"
__credits__ = ["Miguel Ponce de Leon"]
__license__ = "GPL 3.0"
__version__ = "0.1.0"
__maintainer__ = "Miguel Ponce de Leon"
__email__ = "miguel.ponce@bsc.es"
__status__ = "dev"


class Metadata(object):
    def __init__(self, tree):

        root = tree.getroot()
        metadata_node = root.find("metadata")

        node = metadata_node.find("current_time")
        self._current_time = int(float(node.text))
        self._time_units = node.attrib['units']
        
        node = metadata_node.find("current_runtime")
        self._current_runtime = float(node.text)
        self._runtime_units = node.attrib['units']

        micro_env_node = root.find("microenvironment")
        node = micro_env_node.find("domain")
        node = node.find("mesh")
        self._spatial_units = node.attrib['units']
    
    @property
    def current_time(self):
        return self._current_runtime

    @property
    def time_units(self):
        return self._time_units
    
    @property
    def current_runtime(self):
        return self._current_runtime
    
    @property
    def runtime_units(self):
        return self._runtime_units
    
    @property
    def spatial_units(self):
        return self._spatial_units


class MultiCellDS(object):
    
    def __init__(self, output_folder="./", xml_fname="initial.xml", sep="_"):
        
        
        self._param_folder = os.path.join(os.path.dirname(__file__), "params")
        self._globing = os.path.join(output_folder, "output*.xml")
        
        self._separator = sep
        self._output_folder = output_folder

        self._phases_dict = default_phases_dict
        self._phase_grouping = default_phase_grouping

        xml_fname = os.path.join(output_folder, xml_fname)
        self._tree = ET.parse(xml_fname)

        self._metadata = Metadata(self._tree)
        self._cell_columns = self._get_cell_columns()
        self._microenvironment_columns = self._get_microenvironment_columns()


    def _get_time_units(self):
        root = self._tree.getroot()
        node = root.find("metadata")
        node = node.find("current_time")
        units = node.attrib['units']
        return units

    def _get_cell_info_recursive(self, node):

        childs = [i for i in node]
        if len(childs) == 0:
            return None
        for child in childs:
            if child.tag != "simplified_data":
                continue
            if child.attrib["source"] == "PhysiCell":
                return child
    
        return self._get_cell_info_recursive(childs[0])
    
    def _get_cell_columns(self):
        root = self._tree.getroot()
        node = root.find("cellular_information")
        
        node = self._get_cell_info_recursive(node)
        cellular_info_node = node.find("labels")
        
        cell_columns = []
        for child in cellular_info_node:
            column = child.text
            index = child.attrib["index"]
            size = int(child.attrib["size"])
            if size == 1:
                cell_columns.append(column)
                continue
            for i, v in enumerate(['x', 'y', 'z']):
                cell_columns.append(v + self._separator + column)
                if i == size: 
                    break
                
                
        return cell_columns

    def _get_microenvironment_columns(self):
        root = self._tree.getroot()
        node = root.find("microenvironment")
        node = node.find("domain")
        node = node.find("variables")
        
        columns = []
        for child in node.findall("variable"):
            name = child.attrib['name']
            units = child.attrib['units']
            ID = child.attrib['ID']
            columns.append((name, units, ID))
            
        return columns

    @property
    def current_time(self):
        return self._metadata.current_runtime

    @property
    def time_units(self):
        return self._metadata.time_units
    
    @property
    def current_runtime(self):
        return self._metadata.current_runtime
    
    @property
    def runtime_units(self):
        return self._metadata.runtime_units
    
    @property
    def spatial_units(self):
        return self._metadata.spatial_units

    @property
    def cell_columns(self):
        return self._cell_columns

    @property
    def microenvironment_columns(self):
        return self._microenvironment_columns

    @property
    def phases_dict(self):
        return self._phases_dict

    @property
    def phase_grouping(self):
        return self._phase_grouping

    def _read_matlab_mat(self, fname, column):
        stru = loadmat(fname)
        data = stru[column]
        return data

    def get_time(self, tree):
        root = tree.getroot()
        node = root.find("metadata")
        node = node.find("current_time")
        time = int(float(node.text))
        return time

    def cells_file_count(self):
        return len(glob.glob(self._globing))

    def get_cells_fname(self, tree):
        root = tree.getroot()
        node = root.find("cellular_information")
        node = self._get_cell_info_recursive(node)
        node = node.find("filename")
        return node.text

    def get_cells_matrix(self, tree):
        matfile = self.get_cells_fname(tree)
        matfile = os.path.join(self._output_folder, matfile)
        data = self._read_matlab_mat(matfile, "cells")
        return data.T

    def cells_as_matrix_iterator(self):
        xml_list = sorted(glob.glob(self._globing))
        for xml_fname in xml_list:
            tree = ET.parse(xml_fname)
            cell_matrix = self.get_cells_matrix(tree)
            
            time = self.get_time(tree)
            yield (time, cell_matrix)

    def cells_as_frames_iterator(self):
        xml_list = sorted(glob.glob(self._globing))
        for xml_fname in xml_list:
            tree = ET.parse(xml_fname)
            cell_matrix = self.get_cells_matrix(tree)
            
            df = pd.DataFrame(cell_matrix, columns=self._cell_columns)
            df = df.set_index("ID")
        
            time = self.get_time(tree)
            yield (time, df)
  
    def get_microenvironment_fname(self, tree):
        root = tree.getroot()
        node = root.find("microenvironment")
        node = node.find("domain")
        node = node.find("data")
        node = node.find("filename")
        return node.text

    def get_microenvironment_matrix(self, tree):
        matfile = self.get_microenvironment_fname(tree)
        matfile = os.path.join(self._output_folder, matfile)
        data = self._read_matlab_mat(matfile, "multiscale_microenvironment")
        return data
        
    def microenvironment_as_matrix_iterator(self):
        xml_list = sorted(glob.glob(self._globing))
        for xml_fname in xml_list:
            tree = ET.parse(xml_fname)
            microenv_matrix = self.get_microenvironment_matrix(tree)
            time = self.get_time(tree)
            yield (time, microenv_matrix)

    def get_cells_summary_frame(self, phase_col="current_phase"):

        cell_phases = list(set(self.phase_grouping.values()))
        num_of_files = self.cells_file_count()

        # Initializing a Pandas Databrafe to store the data
        index = range(num_of_files)
        columns = ["time"] + cell_phases
        df_time_course = pd.DataFrame(columns=columns, dtype=int, index=index, data=0)
        
        for i, (time, df) in enumerate(self.cells_as_frames_iterator()):
            df_time_course.iloc[i, 0] = time

            # Rename the phases integer codes using the phases_dict as the mapping
            s = df[phase_col]
            s.replace(to_replace=self.phases_dict, value=None, inplace=True)

            # Count the number of cells in each phase
            counts = s.value_counts()

            # group the previous phases count into the three general classes:
            # Alive, Apoptotic, Necrotic
            for k, v in counts.to_dict().items():
                if k not in self.phase_grouping:
                    continue
                df_time_course.loc[i, self.phase_grouping[k]] += v
        
        return df_time_course


    def plot_cells(self):
        color_dict = {"alive": "g", "apoptotic": "r", "necrotic":"k"}

        fig, ax = plt.subplots(1, 1, figsize=(6,4), dpi=150)
        # Alive/Apoptotic/Necrotic vs Time
        for k in columns:
            ax.plot(df_time_course.Time, df_time_course[k], "-", c=color_dict[k], label=k)
        # setting axes labels
        ax.set_xlabel("time (min)")
        ax.set_ylabel("Nº of cells")
        # Showing legend
        ax.legend()
        ax.yaxis.grid(True)
        fig.tight_layout()
        # Saving fig
        if save:
            fig.savefig(output_folder + 'cell_vs_time.png')



#!/usr/bin/env python
# coding: utf-8

import re
import os
import sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

modules_path = os.path.dirname(os.path.realpath(__file__))
modules_path = os.path.join(modules_path, 'modules')
sys.path.append(modules_path)

from multicellds import MultiCellDS

def get_timeserie_mean(mcds, filter_alive=True):
    time = []
    values = []
    filter_alive = True
    for t, df in mcds.cells_as_frames_iterator():
        time.append(t)
        df = df.iloc[:,3:]
        if filter_alive:
            mask = df['current_phase'] <= 14
            df = df[mask]
        values.append(df.mean(axis=0).values)

    cell_columns = df.columns.tolist()
    df = pd.DataFrame(values, columns=cell_columns)
    df['time'] = time
    return df[['time'] + cell_columns]


def get_timeserie_density(mcds):
    data = []
    for t,m in mcds.microenvironment_as_matrix_iterator():
        data.append((t, m[5,:].sum()))
    df = pd.DataFrame(data=data, columns=['time', 'tnf'])
    return df

def plot_molecular_model(df_cell_variables, list_of_variables, ax1):

    threshold = 0.5

    for label in list_of_variables:
        y = df_cell_variables[label]
        time = df_cell_variables["time"]
        ax1.plot(time, y, label="% X " + label)

    ax1.set_ylabel("% X")
    ax1.yaxis.grid(True)
    ax1.set_xlim((0,time.values[-1]))
    ax1.set_ylim((0,1))
    # ax1.set_xlabel("time (min)")
    
def plot_cells(df_time_course, color_dict, ax):

    # Alive/Apoptotic/Necrotic vs Time
    for k in color_dict:
        ax.plot(df_time_course.time, df_time_course[k], "-", c=color_dict[k], label=k)
    
    # setting axes labels
    # ax.set_xlabel("time (min)")
    ax.set_ylabel("Nº of cells")
    
    # Showing legend
    ax.legend()
    ax.yaxis.grid(True)

def main():

    color_dict = {"live": "g", "apoptotic": "r", "necrotic":"k"}

    instance_folder = sys.argv[1]

    output_data = instance_folder + 'output/'

    mcds = MultiCellDS(output_folder=output_data)

    df_time_course = mcds.get_cells_summary_frame()
    df_cell_variables = get_timeserie_mean(mcds)
    df_time_tnf = get_timeserie_density(mcds)

    df_time_course.to_csv(instance_folder + "time_course.tsv", sep="\t")
    df_cell_variables.to_csv(instance_folder + "cell_variables.tsv", sep="\t")
    df_time_tnf.to_csv(instance_folder + "tnf_time.tsv", sep="\t")

    fig, axes = plt.subplots(3, 1, figsize=(12,12), dpi=150, sharex=True)
    plot_cells(df_time_course, color_dict, axes[0])
    
    list_of_variables = ['bound_external_TNFR', 'unbound_external_TNFR', 'bound_internal_TNFR']
    plot_molecular_model(df_cell_variables, list_of_variables, axes[1])
    threshold = 0.5
    
    axes[1].hlines(threshold, 0, df_time_course.time.iloc[-1], label="Activation threshold")
    ax2 = axes[1].twinx()
    ax2.plot(df_time_tnf.time, df_time_tnf['tnf'], 'r', label="[TNF]")
    ax2.set_ylabel("[TNF]")
    ax2.set_ylim([0, 1000])
    axes[1].legend(loc="upper left")
    ax2.legend(loc="upper right")

    list_of_variables = ['tnf_node', 'nfkb_node', 'fadd_node']
    plot_molecular_model(df_cell_variables, list_of_variables, axes[2])
    axes[2].set_xlabel("time (min)")
    ax2 = axes[2].twinx()
    ax2.plot(df_time_tnf.time, df_time_tnf['tnf'], 'r', label="[TNF]")
    ax2.set_ylabel("[TNF]")
    ax2.set_ylim([0, 1000])
    axes[2].legend(loc="upper left")
    ax2.legend(loc="upper right")

    fig.tight_layout()
    fig.savefig(instance_folder + 'variables_vs_time.png')





#!/usr/bin/env python3
# coding: utf-8

import os, re, sys
import glob, json
import argparse

import numpy as np
import pandas as pd
from scipy.io import loadmat

import matplotlib.pyplot as plt
import seaborn as sns

from pctk import multicellds 


sns.set(style="ticks", palette="Paired")
sns.set_context('paper')


def create_parser():
    parser = argparse.ArgumentParser(description="Plot total cell grouped as Alive/Necrotic/Apoptotic vs Time")
    
    parser.add_argument("data_folder", action="store", help="folder were the data is stored")
    
    parser.add_argument("--format", action="store", dest="format", choices=("physicell", "physiboss"),
                        help="Format of the input data", default="physicell")

    parser.add_argument("--figout", action="store", dest="fig_fname", default="./cell_vs_time.png",
                        help="File name to save the plot")
                        
    parser.add_argument("--csvout", action="store", dest="csv_fname", default=None,
                        help="File name to store the summary table used for the plot")

    return parser
    

def pb_output_iterator(data_folder, sep=";"):
    globing = os.path.join(data_folder, "cells_[0-9]*.txt")
    for fname in sorted(glob.glob(globing)):
        df = pd.read_csv(fname, sep=sep)
        t = df.Time[0]
        yield (t, df)

def count_pb_files(data_folder):
    globing = os.path.join(data_folder, "cells_[0-9]*.txt")
    return len(glob.glob(globing))

def main():
   
    parser = create_parser()
    args = parser.parse_args()
    
    phases_dict = multicellds.default_phases_dict
    phase_grouping = multicellds.default_phase_grouping
    print(phases_dict)
    # Globing output files according to the output format specified
    if args.format == 'physicell':
        phase_col = "current_phase"
        mcds = multicellds.MultiCellDS(output_folder=args.data_folder)
        df_iterator = mcds.cells_as_frames_iterator()
        num_of_files = mcds.cells_file_count()
    elif args.format == 'physiboss':
        phase_col = "phase"
        df_iterator = pb_output_iterator(args.data_folder)
        num_of_files = count_pb_files(args.data_folder)
    
    # Initializing a Pandas Databrafe to store the data
    columns = ["time", "live", "apoptotic", "necrotic"]
    data = np.zeros((num_of_files, 4), dtype=int)
    df_time_course = pd.DataFrame(columns=columns, data=data)

    print("Reading cell_output files from %i input files from %s" % (num_of_files, args.data_folder))
    # Iterating over all cell_output files
    for i, (t, df) in enumerate(df_iterator):
        print("\tProcessing time step: %.0f" % t)

        # Rename the phases integer codes using the phases_dict as the mapping
        s = df[phase_col]
        s.replace(to_replace=phases_dict, inplace=True)
        
        # Count the number of cells in each phase
        counts = s.value_counts()
    
        df_time_course.loc[i, 'time'] = t
        # group the previous phases count into the three general classes:
        # Alive, Apoptotic, Necrotic
        for k, v in counts.to_dict().items():
            if k not in phase_grouping:
                continue
            df_time_course.loc[i, phase_grouping[k]] += v
            
    
    print("Finish processing files")    
    # Set time column as the dataframe index
    sns.set_context('paper')
    patch_color = "lightgrey"
    
    print("Creating figure")
    # Create a figure
    fig, ax = plt.subplots(1, 1, figsize=(8,3), dpi=300)
    
    # plot Alive vs Time
    curve_params = {}
    curve_params['live'] = {'color': '#75db75', 'label': 'Alive'}
    curve_params['apoptotic'] = {'color': '#ef4242', 'label': 'Apoptotic'}
    curve_params['necrotic'] = {'color':'#97723d', 'label': 'Necrotic'}
    line_width = 2.
    for k,pdict in curve_params.items():
        c = pdict['color']
        l = pdict['label']
        ax.plot(df_time_course.time, df_time_course[k], "-", c=c, label=l, linewidth=line_width)
    
    # setting axes labels
    ax.set_xlabel("Time (min)")
    ax.set_ylabel("Nº of cells")

    ax.tick_params(axis='x', labelsize=12)
    ax.tick_params(axis='y', labelsize=12)
    
    # Showing legend
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.yaxis.grid(linestyle='dotted')
    ax.legend()
    
    # Saving fig
    fig.tight_layout()
    fig.savefig(args.fig_fname)
    print("Saving fig as %s" % args.fig_fname)
    
    if args.csv_fname:
        df_time_course.to_csv(args.csv_fname, sep="\t")
        print("Saving csv as %s" % args.csv_fname)


if __name__ == '__main__':
    main()




import setuptools
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()



setuptools.setup(
    name="pctk",
    version="0.1.8",
    author="Miguel Ponce de Leon",
    author_email="miguel.ponce@bsc.es",
    description="PhysiCell ToolKit: a tool for handling MultiCellDS output from PhysiCell and PhysiBoSS simulations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PhysiBoSS/pctk",
    packages=setuptools.find_packages(where="src"   ),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    # install_requires=["numpy>=1.22.4", "scipy>=1.5.1", "matplotlib>=3.5.2",  "pandas>=1.4.2",  "seaborn>=0.11.2"],
    setup_requires=['numpy', 'scipy', 'matplotlib',  'pandas',  'seaborn'],
    install_requires=['numpy', 'scipy', 'matplotlib',  'pandas',  'seaborn'],
    entry_points={
        'console_scripts': ['plot-time-course=pctk.cmds.plot_time_course:main',
                            'write-pov-files=pctk.cmds.write_pov_files:main',
                            ]
        
    }
)



import re
import os
import shutil
import argparse
import numpy as np


import multiprocessing as mp
from pctk.povwriter import POVWriter

def create_parser():

    povray_link = "http://www.povray.org/download/"

    parser = argparse.ArgumentParser(description="Render a bunch of PhysiCell output file into povray files")
    
    parser.add_argument("xml_config", action="store", help="XML configuration file")

    parser.add_argument("--idxs", action="store", dest="strn_idxs", 
                        help="String specifing the indexes of the output files. \
                              The supported options include: \
                              - slices: 1:10:1\
                              - indexes: 1,2,5,10\
                              - all (use glob)", 
                        default="")
    
    parser.add_argument("--format", action="store", dest="format", choices=("physicell", "physiboss"),
                        help="Format of the input data", default="physicell")

    parser.add_argument("--render",  action='store_true',
                        help="Render the .pov files into .png. Requires povray ({povray_link})")

    parser.add_argument("--width", action="store", dest="width", type=int, default=2160, 
                        help="Width for povray rendered image")

    parser.add_argument("--height", action="store", dest="height", type=int, default=2160, 
                        help="Heigh for povray rendered image")

    parser.add_argument("--cpus", action="store", dest="cpus", type=int, default=4, 
                        help="Total CPUs availabile to run in parallel using multiprocessing")

    parser.add_argument("--create-config", dest="create_default_config", default=False,
                        help="Output format")

    return parser


def check_povray(exec='povray'):
    global povray_path
    povray_path = shutil.which(exec)
    return (povray_path is not None)


def local_write_pov_file(writer, idx):
    fname = writer.write_pov_file(idx)
    return fname

def render_to_png(pov_fname, width, height):
    cmd_line = f"{povray_path} -h{height} -w{width} -a {pov_fname}"
    exit_flag = os.system(cmd_line)
    if exit_flag != 0:
        print(f"Somthing went wrong running povray {cmd_line}. Command finished with exit flag {exit_flag}")

    png_fname = pov_fname[0:-4] + ".png"
    return png_fname

def animate_pngs(png_files):
    pass


povray_path = None


def main():
    parser = create_parser()
    args = parser.parse_args()

    if args.render:
        assert check_povray()
    
    # Loadgin XML configuration 
    pov_writer = POVWriter(args.xml_config, format=args.format)

    num_of_jobs = args.cpus

    index_list = []    
    pattern_slices = re.compile("(\d*):(\d*):(\d*)")
    pattern_indexes = re.compile("(\d+)(,\d+)*")
    pattern_all = re.compile("^all$")
    
    match = re.search(pattern_slices, args.strn_idxs)
    if match:
        from_idx = int(match.group(1))
        to_idx = int(match.group(2))
        inc = int(match.group(3))
        index_list = list(range(from_idx, to_idx, inc))
    elif re.search(pattern_indexes, args.strn_idxs):
        index_list = [int(i) for i in args.strn_idxs.split(",")]
    elif re.search(pattern_all, args.strn_idxs):
        pass
    else:
        index_list = [pov_writer.options.time_index]
    
    print(f"Start processing  {num_of_jobs} cpus")
    if num_of_jobs > 1:

        pool = mp.Pool(num_of_jobs)
        pov_files = [pool.apply(local_write_pov_file, args=((pov_writer,idx)))
                                        for idx in index_list]
        pool.close()

        print("Finished!")
        if args.render:
            png_files = []
            for fname in pov_files:
                png = render_to_png(fname, args.width, args.height)
                png_files.append(png)
            
            pool.close()
            #animate_pngs(png_files)
        
    else:
        for idx in index_list:
            pov_writer.write_pov_file(idx)


if __name__ == '__main__':
    main()


import os, sys
import glob
from PIL import Image
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPDF, renderPM



output_dir = sys.argv[1]

# Create the frames

patterns = os.path.join(output_dir, "*.svg")
globbing = sorted(glob.glob(patterns))

frames = []
for i in globbing:
    print("Processin image %s" % i)
    drawing = svg2rlg(i)
    new_frame = renderPM.drawToPIL(drawing, dpi=200)
    frames.append(new_frame)

# Save into a GIF file that loops forever
frames[0].save('animation.gif', format='GIF', 
                append_images=frames[1:],
                save_all=True,
                duration=150, loop=0, quality=200)


#!/usr/bin/env python3
# coding: utf-8
import os
import numpy as np
from math import pi, sin, cos

import xml.etree.ElementTree as ET

from pctk.config import phase_grouping 
from pctk.config import phases_dict
from pctk.config import default_pov_colors



__author__ = "Miguel Ponce de Leon"
__copyright__ = "Copyright 2020, PhysiCell ToolsKit project"
__credits__ = ["Miguel Ponce de Leon"]
__license__ = "GPL 3.0"
__version__ = "0.1.0"
__maintainer__ = "Miguel Ponce de Leon"
__email__ = "miguel.ponce@bsc.es"
__status__ = "dev"


class InvalidFormatException(Exception):
    def __init__(self, format):
        self.format = format
        self.message = f"Invalid format {format}. Format must be physicell or physiboss"
        super().__init__(self.message)


class Clipping_Plane:

    def __init__(self):

        self.coefficients = np.zeros(3)

        self.normal = np.array([0,-1,0])
        self.point_on_plane = np.zeros(3)
        self.coefficients = np.zeros(4)
        
        for i in range(3):
            self.coefficients[i] = self.normal[i]
        for i in range(3):
            self.coefficients[3] -= ( self.normal[i]*self.point_on_plane[i] )

    def coefficients_to_normal_point(self):
        for i in range(3):
            self.normal[i] = self.coefficients[i]
	
        self.normal = self.normal / np.sqrt(np.sum(self.normal**2))
        self.point_on_plane = self.normal
        self.point_on_plane *= -1.0
        self.point_on_plane *= self.coefficients[3]

    def signed_distance_to_plane(self, point):
        assert len(point) == 3
        out = self.coefficients[3]
        for i in range(3):
            out += point[i] * self.coefficients[i]
        return out


class PC_Options:
    def __init__(self, folder="sample", filebase="output", format="physicell",
                 time_index=0, camera_distance=1500, nuclear_offset=0.1, cell_bound=750, 
                 threads=1, use_standard_colors=True):

        self.folder = folder
        self.filebase = filebase
        self.format = format
        self.time_index = time_index
        self.filename = self.create_file_name(time_index)
        
        self.nuclear_offset = nuclear_offset
        self.cell_bound = cell_bound
        self.threads = threads
        self.use_standard_colors = use_standard_colors

        self.columns_index_dict = {}
        if self.format == 'physicell':
            self.columns_index_dict = { "cyto_volume": 4, "cell_type": 5, "phase": 7, "nuc_volume": 9}
        elif self.format == 'physiboss':
            self.columns_index_dict = { "cyto_radius": 4, "cell_type": 11, "phase": 13, "nuc_radius": 6}
        else:
            raise InvalidFormatException(self.format)

    def create_file_name(self, index):
        if self.format == 'physicell':
            fname  = "{base_name}{:08}_cells_physicell.mat".format(index, base_name=self.filebase)
        elif self.format == 'physiboss':
            fname  = "cells_{:05}.txt".format(index)
        else:
            fname = None
        fname = os.path.join(self.folder, fname)
        return fname


class POV_Options():
    
    def __init__(self, x_size=2000, y_size=2000, z_size=2000):
        
        self.max_trace_level = 1
        self.assumed_gamma = 1.0
    
        self.domain_center = np.array([0,0,0])
        self.domain_size = np.array([x_size, y_size, z_size])
        
        self.background = np.array([1,1,1])

        self.camera_theta = 5*pi/4.0
        self.camera_phi = pi/3.0
        self.camera_distance = ( self.domain_size[0] + self.domain_size[1] + self.domain_size[2] ) * 1.25 / 3.0
        self.camera_position = np.array([0,0,0])
        self.set_camera_from_spherical_location()
        
        self.camera_look_at = self.domain_center.copy()
        self.camera_right = np.array([-1,0,0]) # computer graphics are goofy. switch to usual coordinate system
        self.camera_up = np.array([0,0,1])
        self.camera_sky = np.array([0,0,1])
        
        self.light_position = np.array([0,0,0])
    
        self.light_position[0] = self.domain_center[0] - 0.5 * self.domain_size[0]; 
        self.light_position[1] = self.domain_center[1] - 0.5 * self.domain_size[1]; 
        self.light_position[2] = self.domain_center[2] + 1.5 * self.domain_size[2]; 
    
        self.light_rgb = 0.75; 
        self.light_fade_distance = np.linalg.norm( self.domain_size ); 
        self.light_fade_power = 2; 
        
        self.no_shadow = False
        self.no_reflection = False
        
        self.clipping_planes = []
        
    def set_camera_from_spherical_location(self):
        self.camera_position[0] = self.domain_center[0] + self.camera_distance * cos(self.camera_theta)*sin(self.camera_phi) 
        self.camera_position[1] = self.domain_center[1] + self.camera_distance * sin(self.camera_theta)*sin(self.camera_phi)
        self.camera_position[2] = self.domain_center[2] + self.camera_distance * cos(self.camera_phi)


class POVWriter_config():
    def __init__(self):
        self.options = PC_Options()
        self.pov_options = POV_Options()
        self.cell_color_definitions = {0: default_pov_colors}


    # This method parse XML povwriter config file
    def load_config_file(self, fname):
        
        tree = ET.parse(fname)
        root = tree.getroot()
        
        # Parsing save node
        node = root.find("save")
        self.options.folder = node.find("folder").text
        self.options.filebase = node.find("filebase").text
        self.options.time_index = int(node.find("time_index").text)
        self.options.filename = self.options.create_file_name(self.options.time_index)

        # Parsing options node
        node = root.find("options")
        self.options.nuclear_offset = float(node.find("nuclear_offset").text)
        self.options.cell_bound = float(node.find("cell_bound").text)
        self.options.threads = int(node.find("threads").text)

        use_standard_colors = node.find("use_standard_colors").text == 'true'
        if use_standard_colors:
            pigment_and_finish_function = standard_pigment_and_finish_function
        else:
            pigment_and_finish_function = my_pigment_and_finish_function

        # Parsing camera node
        node = root.find("camera")
        self.pov_options.camera_distance = float(node.find("distance_from_origin").text)
        self.pov_options.camera_theta = float(node.find("xy_angle").text)
        self.pov_options.camera_phi = float(node.find("yz_angle").text)
        
        self.pov_options.set_camera_from_spherical_location()
        self.pov_options.light_position[0] *= 0.5
        
        # Parsing clipping_planes node
        node = root.find("clipping_planes")
        for cp_node in node:
            cp = Clipping_Plane()
            vec = np.array([float(i) for i in cp_node.text.split(",")])
            cp.coefficients = vec
            cp.coefficients_to_normal_point()
            self.pov_options.clipping_planes.append(cp)
        print("Found %i clipping planes" % len(self.pov_options.clipping_planes))

        # Parsing cell_color_definitions node
        node = root.find("cell_color_definitions")
        for cell_cd_node in node:
            cell_type = int(cell_cd_node.attrib['type'])
            phase_color_dict = {}
            for phase_node in cell_cd_node:
                phase_name = phase_node.tag
                phase_color_dict[phase_name] = {}
                for comp_node in phase_node:
                    comp = comp_node.tag
                    color_array = np.array([float(i) for i in comp_node.text.split(",")])
                    phase_color_dict[phase_name][comp] = color_array
                
            self.cell_color_definitions[cell_type] = phase_color_dict
        
        print("Found %i cell color definitions ... " % len(self.cell_color_definitions) )


class POVWriter():
    def __init__(self, xml_config, format='physicell'):

        self._config = POVWriter_config()
        self._config.load_config_file(xml_config)
        self.format = format
    
    @property
    def pov_options(self):
        return self._config.pov_options

    @property
    def options(self):
        return self._config.options

    @property
    def threads(self):
        return self.options.threads
    
    @property
    def format(self):
        return self.options.format

    @property
    def cell_color_definitions(self):
        return self._config.cell_color_definitions

    @format.setter
    def format(self, format):
        self.options.format = format

    def _write_pov_header(self, fh):
        fh.write("#include \"colors.inc\"\n")
        fh.write("#include \"shapes.inc\" \n\n")
                
        fh.write("global_settings {\n")
        fh.write("  max_trace_level %i\n" % (self.pov_options.max_trace_level) )
        fh.write("  assumed_gamma %.4f\n" % (self.pov_options.assumed_gamma) )
        fh.write("}\n\n")
        fh.write("background {\n")
        fh.write("  color rgb <%i,%i,%i>\n" % (self.pov_options.background[0], 
                                               self.pov_options.background[1], 
                                               self.pov_options.background[2]) ) 
        fh.write("}\n\n")
        fh.write("camera {\n")
        fh.write("  location <%.3f,%.3f,%.3f>\n" % (self.pov_options.camera_position[0],
                                                    self.pov_options.camera_position[1],
                                                    self.pov_options.camera_position[2]) ) 
        fh.write("  right x\n")
        fh.write("  look_at <%i,%i,%i>\n" % (self.pov_options.camera_look_at[0], 
                                             self.pov_options.camera_look_at[1],
                                             self.pov_options.camera_look_at[2]) )  
        
        fh.write("  right <%i,%i,%i>\n" % (self.pov_options.camera_right[0], 
                                           self.pov_options.camera_right[1],
                                           self.pov_options.camera_right[2]) )  
        
        fh.write("  up <%i,%i,%i>\n" % (self.pov_options.camera_up[0], 
                                        self.pov_options.camera_up[1],
                                        self.pov_options.camera_up[2]) )  

        fh.write("  sky <%i,%i,%i>\n" % (self.pov_options.camera_sky[0],
                                         self.pov_options.camera_sky[1],
                                         self.pov_options.camera_sky[2]) )  
        fh.write(" }\n\n")
        fh.write("light_source {\n")
        fh.write("  <%i,%i,%i>\n" % (self.pov_options.light_position[0],
                                     self.pov_options.light_position[1],
                                     self.pov_options.light_position[2]) )  
        
        fh.write("  color rgb %.2f\n" % (self.pov_options.light_rgb) )
        fh.write("  fade_distance %.2f\n" % (self.pov_options.light_fade_distance) )
        fh.write("  fade_power %i\n" % (self.pov_options.light_fade_power) )
        fh.write("}\n\n")

    def _write_all_cells(self, fh, cells):
        bound = self.options.cell_bound
        for i,row in enumerate(cells):
            if row[1] < -bound or bound < row[1]:
                continue
            if row[2] < -bound or bound < row[2]:
                continue
            if row[3] < -bound or bound < row[3]:
                continue
            
            self._write_cell(fh, row)

    def _write_cell(self, fh, row):
        
        cell_type_index = self.options.columns_index_dict["cell_type"]
        phase_idx = self.options.columns_index_dict["phase"]
        
        cell_type = int(row[cell_type_index])
        current_phase_idx = int(row[phase_idx])
        
        current_phase_name = phases_dict[current_phase_idx]
        current_phase_name = phase_grouping[current_phase_name]
        colors = self.cell_color_definitions[cell_type][current_phase_name]

        center = row[1:4]

        # cytoplasm radius
        if self.options.format == 'physiboss':
            col_idx = self.options.columns_index_dict["cyto_radius"]
            radius = row[col_idx]
        
        elif self.options.format == 'physicell':
            col_idx = self.options.columns_index_dict["cyto_volume"]
            radius = pow( 3/(4*pi) * abs(row[col_idx]), (1./3.) )
        
        render = len(self.pov_options.clipping_planes) == 0
        intersect = False
        for i,cp in enumerate(self.pov_options.clipping_planes):
            dist = cp.signed_distance_to_plane(center)
            if dist <= -radius:
                render = True
            if -radius < dist <= radius:
                render = True
                intersect = True
        
        if intersect:
            fh.write("intersection{ \n")
        if render:
            if intersect:
                fh.write("union{ \n")
                for cp in self.pov_options.clipping_planes:
                    fh.write("plane{<")
                    fh.write("%.3f," % cp.coefficients[0])
                    fh.write("%.3f," % cp.coefficients[1])
                    fh.write("%.3f>," % cp.coefficients[2])
                    fh.write("%.3f\n" % cp.coefficients[3])
                    fh.write(" pigment {color rgb<")
                    fh.write("%.3f," % colors["cytoplasm"][0])
                    fh.write("%.3f," % colors["cytoplasm"][1])
                    fh.write("%.3f>}\n" % colors["cytoplasm"][2])
                    fh.write(" finish {ambient %.3f " % colors["finish"][0] )
                    fh.write("diffuse %.3f " % colors["finish"][1] )
                    fh.write("specular %.3f } }\n" % colors["finish"][2] )
                fh.write("} \n")
            self._write_pov_sphere(fh, center, radius, colors["cytoplasm"], colors["finish"])
        if intersect:
            fh.write("}\n")    


        # nuclear radius
        col_idx = self.options.columns_index_dict["nuc_volume"]
        radius = pow( 3/(4*pi) * abs(row[col_idx]), (1./3.) )
        if self.options.format == 'physiboss':
            col_idx = self.options.columns_index_dict["nuc_radius"]
            radius = row[col_idx]

        
        render = len(self.pov_options.clipping_planes) == 0
        intersect = False
        for i,cp in enumerate(self.pov_options.clipping_planes):
            dist = cp.signed_distance_to_plane(center)        
            if dist <= -(radius + self.options.nuclear_offset):
                render = True
            if -(radius + self.options.nuclear_offset) < dist <= (radius + self.options.nuclear_offset):
                render = True
                intersect = True
        
        if intersect > 0:
            fh.write("intersection{ \n")
        if render:
            if intersect:
                fh.write("union{ \n")
                for cp in self.pov_options.clipping_planes:
                    fh.write("plane{<")
                    fh.write("%.3f," % cp.coefficients[0])
                    fh.write("%.3f," % cp.coefficients[1])
                    fh.write("%.3f>," % cp.coefficients[2])
                    fh.write("%.3f\n" % (cp.coefficients[3]+self.options.nuclear_offset))
                    fh.write(" pigment {color rgb<")
                    fh.write("%.3f," % colors["nuclear"][0])
                    fh.write("%.3f," % colors["nuclear"][1])
                    fh.write("%.3f>}\n" % colors["nuclear"][2])
                    fh.write(" finish {ambient %.3f " % colors["finish"][0] )
                    fh.write("diffuse %.3f " % colors["finish"][1] )
                    fh.write("specular %.3f } }\n" % colors["finish"][2] )
                fh.write("} \n")
            self._write_pov_sphere(fh, center, radius, colors["nuclear"], colors["finish"], no_shadow=True)
        if intersect:
            fh.write("}\n")

    def _write_pov_sphere(self, fh, center, radius, pigment, finish, no_shadow=False, no_reflection=False):
        fh.write("sphere\n{\n")
        fh.write(" <%.4f,%.4f,%.4f>, %.4f" % (center[0], center[1], center[2], radius) )
        fh.write(" pigment {color rgb<%.2f,%.2f,%.2f>}\n" % (pigment[0],pigment[1],pigment[2]) )
        fh.write(" finish {ambient %.2f diffuse %.2f specular %.2f}\n" % (finish[0],finish[1],finish[2]) )
        if no_shadow:
            fh.write(" no_shadow ")
        if no_reflection:
            fh.write(" no_reflection ")
        fh.write("}\n")

    def write_pov_file(self, idx):

        fname = self.options.create_file_name(idx)
        print("Processing file ", fname)
        
        mat = self.read_cells_file(fname)
        print("Matrix size: %i x %i " % mat.shape)
        
        pov_fname = fname[:-4] + ".pov"
        with open(pov_fname, 'w') as fh:
            print("Creating file %s for output ... " % pov_fname)
            self._write_pov_header(fh)       
            print("Writing %i cells ... " % mat.shape[0])
            self._write_all_cells(fh, mat)
            return pov_fname
        
        return None

    def read_cells_file(self, fname):
        if self.format == 'physicell':
            from scipy.io import loadmat
            mat = loadmat(fname)['cells'].T
            return mat
        elif self.format == 'physiboss':
            data = []
            with open(fname) as fh:
                header = next(fh)
                for row in fh:
                    data.append([float(i) for i in row.split(";")])
                data = np.array(data)
                # Removing time storing column
                data = data[:,1:]
            return data
        else:
            raise InvalidFormatException(self.format)



def standard_pigment_and_finish_function():
    pass

def my_pigment_and_finish_function():
    pass



phase_grouping = { 
    "Ki67_positive_premitotic": "live",  
    "Ki67_positive_postmitotic": "live", 
    "Ki67_positive": "live", 
    "Ki67_negative": "live", 
    "G0G1_phase": "live", 
    "G0_phase": "live", 
    "G1_phase": "live", 
    "G1a_phase": "live", 
    "G1b_phase": "live", 
    "G1c_phase": "live", 
    "S_phase": "live", 
    "G2M_phase": "live", 
    "G2_phase": "live", 
    "M_phase": "live", 
    "live": "live", 
    "apoptotic": "apoptotic",  
    "necrotic_lysed": "necrotic", 
    "necrotic_swelling": "necrotic"
    }

phases_dict = {
    0: "Ki67_positive_premitotic",
    1: "Ki67_positive_postmitotic",
    2: "Ki67_positive",
    3: "Ki67_negative",
    4: "G0G1_phase",
    5: "G0_phase",
    6: "G1_phase",
    7: "G1a_phase",
    8: "G1b_phase",
    9: "G1c_phase",
    10: "S_phase",
    11: "G2M_phase",
    12: "G2_phase",
    13: "M_phase",
    14: "live",
    100: "apoptotic",
    101: "necrotic_swelling",
    102: "necrotic_lysed",
    103: "necrotic",
    104: "debris"
    }


default_pov_colors = { 
    'live':{
            'cytoplasm': [.25, 1, .25],
            'nuclear':   [0.03, 0.125, 0],
            'finish':    [0.05, 1, 0.1]
        }, 
    'apoptotic': {
            'cytoplasm': [ 1, 0, 0],
            'nuclear':   [0.125, 0, 0],
            'finish':    [0.05, 1, 0.1]

         }, 
    'necrotic': {
            'cytoplasm': [1, 0.5412, 0.1490],
            'nuclear':   [0.125, 0.06765, 0.018625],
            'finish':    [0.01, 0.5, 0.1]
        } 
    }
